import abc
import numpy as np
import tensorflow as tf
from refactored_DC.settings import CLASSIFICATION_KEY, REGRESSION_KEY, EPSILON

class DataLoader(object, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def next_batch(self, batch_size):
        raise NotImplementedError('users must define next_batch to create a data loader')
    
    @abc.abstractmethod
    def get_epochs(self, batch_size):
        raise NotImplementedError('users must define get_epochs to create a data loader')
    
    @abc.abstractmethod
    def reset_cursor(self, batch_size):
        raise NotImplementedError('users must define reset_cursor to create a data loader')

class AugmentedDataLoader(object, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def next_batch(self, batch_size):
        raise NotImplementedError('users must define next_batch to create an augmented data loader')
    
    @abc.abstractmethod
    def activate_augmentation(self):
        raise NotImplementedError('users must define activate_augmentation to create an augmented data loader')

    @abc.abstractmethod
    def deactivate_augmentation(self):
        raise NotImplementedError('users must define activate_augmentation to create an augmented data loader')
    
    @abc.abstractmethod
    def get_epochs(self, batch_size):
        raise NotImplementedError('users must define get_epochs to create a data loader')
    
    @abc.abstractmethod
    def reset_cursor(self, batch_size):
        raise NotImplementedError('users must define reset_cursor to create a data loader')

class DataLoaderFromArrays(DataLoader):

    def __init__(self, features, targets, homogeneous=True, 
                problem_type=CLASSIFICATION_KEY, shuffle=True, 
                one_hot=True, normalization=True, target_scaling=True, target_range=[0.0,1.0]):
        self.curr_pos = 0
        self.target_divider = None
        self.rows_count = features.shape[0]
        self.homogeneous = homogeneous
        self.to_shuffle = shuffle
        self.problem_type = problem_type
        self.normalization = normalization
        self.target_scaling = target_scaling
        self.target_range = target_range
        self.one_hot = one_hot
        self.features = self.normalize(features)
        self.targets = self.preprocess(targets)
        
    def next_batch(self, batch_size):
        if self.curr_pos+batch_size >= self.rows_count:
            batch = self.features[self.curr_pos:self.rows_count], self.targets[self.curr_pos:self.rows_count]
            self.curr_pos = 0
            if self.to_shuffle:
                self.shuffle()
            return batch
        batch = self.features[self.curr_pos:self.curr_pos+batch_size], self.targets[self.curr_pos:self.curr_pos+batch_size]
        self.curr_pos += batch_size
        return batch
    
    def shuffle(self):
        indices = np.arange(0, self.rows_count) 
        np.random.shuffle(indices) 
        self.features = self.features[indices]
        self.targets = self.targets[indices]

    def deprocess(self, perf):
        if self.problem_type == REGRESSION_KEY and self.target_scaling:
            deprocessed_perf = (perf/(self.target_range[1]-self.target_range[0]))*self.target_divider
            return deprocessed_perf
        else:
            return perf

    def preprocess(self, targets):
        if self.problem_type == REGRESSION_KEY and self.target_scaling:
            mi = targets.min(axis=0)
            divider = targets.max(axis=0) - mi
            if isinstance(divider, np.ndarray):
                divider[divider==0.0] = EPSILON
            else:
                divider = EPSILON if divider==0.0 else divider
            targets = self.target_range[0] + np.float32((targets-mi)/divider)*(self.target_range[1]-self.target_range[0])
            self.target_divider = divider
        elif self.one_hot:
            onehot_targets = np.zeros((self.rows_count, targets.max()+1))
            onehot_targets[np.arange(self.rows_count),targets] = 1
            targets = onehot_targets
        return targets

    def normalize(self, data):
        if not self.normalization:
            return data
        if self.homogeneous:
            mi = data.min()
            divider = data.max() - mi
            divider = EPSILON if divider==0.0 else divider
        else:
            mi = data.min(axis=0)
            divider = data.max(axis=0) - mi
            if isinstance(divider, np.ndarray):
                divider[divider==0.0] = EPSILON
            else:
                divider = EPSILON if divider==0.0 else divider
        return np.float32((data-mi)/divider)

    def get_epochs(self, batch_size):
        if self.rows_count % batch_size != 0:
            return self.rows_count // batch_size + 1  
        else:
            return self.rows_count // batch_size 
    
    def reset_cursor(self):
        self.curr_pos = 0
